/* Copyright (c) 2002 Bruce Spencer, National Research Council, Canada */

import java.io.*;
import java.util.*;

/**
   <code>BUDCTree</code> is a program to create DefiniteClause facts bottom up.

   @author Bruce Spencer <Bruce.Spencer@nrc.ca>
   @version 
*/
public class BUDCTree{

  static SymbolTable symbolTable = new SymbolTable();
  static DiscTree oldFactDiscTree = new DiscTree(symbolTable);
  static DiscTree newRuleDiscTree = new DiscTree(symbolTable);
  static Heap newFactHeap = new Heap();

  static final boolean TRACE = false;

  public static void main(String args[]) throws org.jdom.JDOMException{


    DCFileParser dcfp = new DCFileParser(symbolTable);
    dcfp.parseDCFile("examples.dc");
    loadDefiniteClauses(dcfp.iterator());

    RuleMLFileParser rfp = new RuleMLFileParser(symbolTable);
    rfp.parseRuleMLFile("discount.ruleml");
    loadDefiniteClauses(rfp.iterator());

    if(TRACE){
      System.out.println("Old Facts:\n" + oldFactDiscTree + "\n");
      System.out.println("New Rules:\n" + newRuleDiscTree + "\n");
      System.out.println("New Facts:" + newFactHeap + "\n");
    }

    runForwardReasoner();

    if(TRACE){
      System.out.println("Final Old Facts\n" + oldFactDiscTree);
    }

    System.out.println("All Known Facts");
    Iterator lit = oldFactDiscTree.allLeavesIterator();
    int j = 0;
    while(lit.hasNext()){
      j++;
      System.out.println(j+": " + ((DefiniteClause) lit.next()) );
    }
  }

  public static void runForwardReasoner(){
    
    while(true){
      if(newFactHeap.isEmpty()) return;
	
      DefiniteClause newFact = 
	(DefiniteClause) newFactHeap.removeSmallest();

      if(forwardSubsumes(oldFactDiscTree, newFact)){
	if(TRACE){
	  System.out.println("newFact " + newFact + 
			     " was forward subsumed");
	}
      }
      else {
	if(TRACE){
	  System.out.println("Selected newFact " + newFact);
	}
	
	DiscTree.UnifiableIterator uif = 
	  newRuleDiscTree.unifiableIterator(newFact,0);
	
	while(uif.hasNext()){
	  DefiniteClause newRule = (DefiniteClause) uif.next();
	  Unifier u = new Unifier(newRule, newFact);
	  if(u.unified){
	    DefiniteClause newResult = u.resolvent();
	    if(TRACE){
	      System.out.println("new resolvent " + newResult);
	    }
	    process(newResult);
	  }
	}
	oldFactDiscTree.insert(newFact, 0);
      }
    }
  }

  public static void process(DefiniteClause result){
  
    if(result.symbols.length == result.symbols[0][1]){
      //if result is a fact
      newFactHeap.add(result);
    }
    else {
      DiscTree.UnifiableIterator uif = 
  	oldFactDiscTree.unifiableIterator(result, 1);
      while(uif.hasNext()){
  	DefiniteClause oldFact = (DefiniteClause) uif.next();
  	Unifier u = new Unifier(result, oldFact);
  	if(u.unified){
  	  DefiniteClause newResult = u.resolvent();
	  if(TRACE){	    
	    System.out.println("new resolvent " + newResult);
	  }
  	  process(newResult);
  	}
      }
      newRuleDiscTree.insert(result,1);
    }
  }

  public static void loadDefiniteClauses(Iterator it){
    while(it.hasNext()){
      DefiniteClause dc = (DefiniteClause) it.next();
      if(dc.symbols.length == dc.symbols[0][1]){
	//then this definite clause is a fact
	//so put it in with the facts
	newFactHeap.add(dc);
      }
      else{
	//else it is a new rule
	newRuleDiscTree.insert(dc, 1);
      }
    }
  }
    

  public static boolean forwardSubsumes
    (DiscTree oldFactDiscTree, DefiniteClause newFact){
    Unifier u1 = new Unifier(newFact, Unifier.GROUNDING_MODE);
    DiscTree.UnifiableIterator uif = 
      oldFactDiscTree.unifiableIterator(newFact, 0);
    while(uif.hasNext()){
      DefiniteClause oldFact = (DefiniteClause) uif.next();
      Unifier u2 = new Unifier(oldFact,0,0,u1,Unifier.ONEWAY_MODE);
      if(u2.unified){
	if(TRACE){
	  System.out.println
	    ("new fact " + newFact + 
	     " subsumed by old Fact " + oldFact);
	}
	return true;
      }
    }
    return false;
  }
}






